#include "config.h"

#include <libguile.h>
#include <guile/gh.h>

#include <GL/glu.h>
#include <guile-gtk.h>

#include "GL-support.h"
#include "fakeGLUtypes.h"

extern sgtk_enum_info sgtk_glu_tess_property_type_info;
extern sgtk_enum_info sgtk_glu_tess_winding_rule_info;
extern sgtk_enum_info sgtk_glu_tess_callback_type_info;
extern sgtk_enum_info sgtk_gl_primitive_type_info;

long		tc_tess;

#ifndef HAVE_GLU_1_2_TESS
void sglu_init_GLU_tess_glue (void)
{
  /* sorry - no tesselators :o( */
}

#else

/*
 *
 *	Tesselators
 */

#define CB_BEGIN		0
#define CB_EDGE_FLAG		1
#define CB_VERTEX		2
#define CB_END			3
#define CB_COMBINE		4
#define CB_ERROR		5

#define	TESS_NOF_CALLBACKS	6

SCM smob_tess_mark (SCM s_tess)
{
  TessWrap	*wptr;

  wptr = SCM_TESS (s_tess);

  scm_gc_mark (wptr -> protect_list);
  scm_gc_mark (wptr -> callbacks);
  scm_gc_mark (wptr -> callbacks_with_data);
  scm_gc_mark (wptr -> data);

  return SCM_BOOL_F;
}

SCM_PROC (s_glu_new_tess, "glu-new-tess", 0, 0, 0, sglu_new_tess);

static SCM
sglu_new_tess (void)
{
  TessWrap	*wptr;

  wptr = (TessWrap *) scm_must_malloc (sizeof (TessWrap), "TessWrap");
  wptr -> tess = gluNewTess ();

  wptr -> protect_list = SCM_EOL;
  wptr -> callbacks = gh_make_vector (SCM_MAKINUM (6), SCM_BOOL_F);
  wptr -> callbacks_with_data = gh_make_vector (SCM_MAKINUM (6), SCM_BOOL_F);
  wptr -> data = SCM_BOOL_F;

  SCM_RETURN_NEWSMOB (tc_tess, wptr);
}

static void*
tess_scm_to_voidp (TessWrap *ptr, SCM data)
{
#if SIZEOF_SCM <= SIZEOF_VOID_P
  if (SCM_NIMP (data))
    ptr -> protect_list = scm_cons (data, ptr -> protect_list);

  return  (void *) data;
#else
  ptr -> protect_list = SCM_CONS (data, ptr->protect_list);
  return SCM_CARLOC (ptr->protect_list);
#endif
}

static SCM
tess_voidp_to_scm (void *ptr)
{
#if SIZEOF_SCM <= SIZEOF_VOID_P
  return (SCM) ptr;
#else
  return *((SCM *) ptr);
#endif
}

SCM_PROC (s_glu_tess_begin_polygon, "glu-tess-begin-polygon", 1, 1, 0, sglu_tess_begin_polygon);

static SCM
sglu_tess_begin_polygon (SCM s_tess, SCM s_data)
{
  TessWrap		*wptr;

  SCM_ASSERT (SCM_TESSP (s_tess), s_tess, SCM_ARG1, s_glu_tess_begin_polygon);

  wptr = SCM_TESS (s_tess);
  SCM_ASSERT (wptr -> data == SCM_BOOL_F, s_tess, SCM_ARG1, s_glu_tess_begin_polygon);

  wptr -> data = (s_data == SCM_UNDEFINED) ? SCM_BOOL_F : s_data;
  gluTessBeginPolygon (wptr->tess, wptr);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glu_tess_vertex, "glu-tess-vertex", 3, 0, 0, sglu_tess_vertex);
static SCM
sglu_tess_vertex (SCM s_tess, SCM s_vertices, SCM s_data)
{
  GLdouble	*vertices;
  GLdouble	vertices_m [3];
  TessWrap	*wptr;

  SCM_ASSERT (SCM_TESSP (s_tess), s_tess, SCM_ARG1, s_glu_tess_vertex);
  vertices = sgl_scm2d_array (s_vertices, vertices_m, 3, TRUE, SCM_ARG2, s_glu_tess_vertex);
  
  wptr = SCM_TESS (s_tess);
  gluTessVertex (wptr->tess, vertices, tess_scm_to_voidp (wptr, s_data));

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glu_tess_end_polygon, "glu-tess-end-polygon", 1,0,0, sglu_tess_end_polygon);
static SCM
sglu_tess_end_polygon (SCM s_tess)
{
  TessWrap	*wptr;

  SCM_ASSERT (SCM_TESSP (s_tess), s_tess, SCM_ARG1, s_glu_tess_vertex);

  wptr = SCM_TESS (s_tess);
  gluTessEndPolygon (wptr->tess);

  /* now protect list can be "donated" to gc */
  /* callback data will not be needed anymore */
  wptr -> protect_list = SCM_EOL; 
  wptr -> data = SCM_BOOL_F;
  
  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glu_tess_property, "glu-tess-property", 3, 0, 0, sglu_tess_property);

static SCM
sglu_tess_property (SCM s_tess, SCM s_which, SCM s_value)
{
  int		 which;
  GLdouble	 value;

  SCM_ASSERT (SCM_TESSP (s_tess), s_tess, SCM_ARG1, s_glu_tess_property);
  which = sgtk_scm2enum (s_which, &sgtk_glu_tess_property_type_info, 
			 SCM_ARG2, (char *) s_glu_tess_property);
  if (scm_number_p (s_value) == SCM_BOOL_T)
    value = gh_scm2double (s_value);
  else
    if (s_value == SCM_BOOL_F || s_value == SCM_BOOL_T)
      value = (s_value == SCM_BOOL_T) ? 1.0 : 0.0;
    else
      value = sgtk_scm2enum (s_value, &sgtk_glu_tess_winding_rule_info, 
			     SCM_ARG3, (char *)  s_glu_tess_property);

  gluTessProperty (SCM_TESS (s_tess) -> tess, which, value);

  return SCM_UNSPECIFIED;
}

SCM_PROC (s_glu_get_tess_property, "glu-get-tess-property", 2, 0, 0, sglu_get_tess_property);

static SCM
sglu_get_tess_property (SCM s_tess, SCM s_which)
{
  int		 which;
  GLdouble	 value;

  SCM_ASSERT (SCM_TESSP (s_tess), s_tess, SCM_ARG1, s_glu_get_tess_property);
  which = sgtk_scm2enum (s_which, &sgtk_glu_tess_property_type_info, 
			 SCM_ARG2, (char *) s_glu_get_tess_property);

  gluGetTessProperty (SCM_TESS (s_tess) -> tess, which, &value);

  switch (which)
    {
    case GLU_TESS_WINDING_RULE:
      return sgtk_enum2scm (value, &sgtk_glu_tess_winding_rule_info);
    case GLU_TESS_BOUNDARY_ONLY:
      return value == 0.0 ? SCM_BOOL_F : SCM_BOOL_T;
    case GLU_TESS_TOLERANCE:
      return gh_double2scm (value);
    default:
      SCM_ASSERT (0, s_which, SCM_ARG2, s_glu_get_tess_property);
    }

  return SCM_UNSPECIFIED;
}

/*
 *
 *		CALLBACKS
 */

static void
sglu_begin_callback (GLenum type, TessWrap *wptr)
{
  SCM	s_type;
  SCM	func, func_data;

  func = SCM_VELTS (wptr->callbacks) [CB_BEGIN];
  func_data = SCM_VELTS (wptr->callbacks_with_data) [CB_BEGIN];

  s_type = sgtk_enum2scm (type, &sgtk_gl_primitive_type_info);

  if (func_data != SCM_BOOL_F)
    gh_call2 (func_data, s_type, wptr -> data);
  else if (func != SCM_BOOL_F)
    gh_call1 (func, s_type);
}

static void
sglu_edge_flag_callback (GLboolean flag, TessWrap *wptr)
{
  SCM	s_flag;
  SCM	func, func_data;

  func = SCM_VELTS (wptr->callbacks) [CB_EDGE_FLAG];
  func_data = SCM_VELTS (wptr->callbacks_with_data) [CB_EDGE_FLAG];

  s_flag = (flag == TRUE) ? SCM_BOOL_T : SCM_BOOL_F;

  if (func_data != SCM_BOOL_F)
    gh_call2 (func_data, s_flag, wptr -> data);
  else if (func != SCM_BOOL_F)
    gh_call1 (func, s_flag);
}

static void
sglu_vertex_callback (void *data, TessWrap *wptr)
{
  SCM	s_data;
  SCM	func, func_data;

  func = SCM_VELTS (wptr->callbacks) [CB_VERTEX];
  func_data = SCM_VELTS (wptr->callbacks_with_data) [CB_VERTEX];

  s_data = tess_voidp_to_scm (data);

  if (func_data != SCM_BOOL_F)
    gh_call2 (func_data, s_data, wptr -> data);
  else if (func != SCM_BOOL_F)
    gh_call1 (func, s_data);
}

static void
sglu_end_callback (TessWrap *wptr)
{
  SCM	func, func_data;

  func = SCM_VELTS (wptr->callbacks) [CB_END];
  func_data = SCM_VELTS (wptr->callbacks_with_data) [CB_END];

  if (func_data != SCM_BOOL_F)
    gh_call1 (func_data, wptr -> data);
  else if (func != SCM_BOOL_F)
    gh_call0 (func);
}

static void
sglu_combine_callback (GLdouble	  coords[3], 
		       void	 *vertex_data [4],
		       GLfloat	  weight [4],
		       void	**out_data,
		       TessWrap	 *wptr)
{
  SCM	func, func_data;
  SCM	s_coords, s_vdata, s_weight, s_outdata;
  
  func = SCM_VELTS (wptr->callbacks) [CB_COMBINE];
  func_data = SCM_VELTS (wptr->callbacks_with_data) [CB_COMBINE];

  s_coords = gh_doubles2scm (coords, 3);
  s_vdata = scm_listify (tess_voidp_to_scm (vertex_data [0]),
			 tess_voidp_to_scm (vertex_data [1]),
			 tess_voidp_to_scm (vertex_data [2]),
			 tess_voidp_to_scm (vertex_data [3]),
			 SCM_UNDEFINED);
  s_weight = scm_listify (scm_make_real	(weight [0]),
			  scm_make_real	(weight [1]),
			  scm_make_real	(weight [2]),
			  scm_make_real	(weight [3]),
			  SCM_UNDEFINED);

  if (func_data != SCM_BOOL_F)
    s_outdata = scm_apply (func_data, s_coords, 
			   scm_cons2 (s_vdata, s_weight, 
				      scm_cons (wptr -> data, scm_listofnull)));
  else if (func != SCM_BOOL_F)
    s_outdata = gh_call3 (func_data, s_coords, s_vdata, s_weight);

  *out_data =  tess_scm_to_voidp (wptr, s_outdata);
}

static void
sglu_error_callback (GLenum errno, TessWrap *wptr)
{
  SCM	func, func_data;
  SCM	s_errno;

  func = SCM_VELTS (wptr->callbacks) [CB_ERROR];
  func_data = SCM_VELTS (wptr->callbacks_with_data) [CB_ERROR];

  /* FIXME */
  s_errno = SCM_MAKINUM (errno);

  if (func_data != SCM_BOOL_F)
    gh_call2 (func_data, s_errno, wptr -> data);
  else if (func != SCM_BOOL_F)
    gh_call1 (func, s_errno);
}

typedef void (*CBFunc)(void);

static CBFunc cb_func_array [TESS_NOF_CALLBACKS] = {
  (CBFunc) sglu_begin_callback,
  (CBFunc) sglu_edge_flag_callback,
  (CBFunc) sglu_vertex_callback, 
  (CBFunc) sglu_end_callback,
  (CBFunc) sglu_combine_callback,
  (CBFunc) sglu_error_callback
};

SCM_PROC (s_glu_tess_callback, "glu-tess-callback", 3, 0, 0, sglu_tess_callback);

static SCM
sglu_tess_callback (SCM s_tess, SCM s_which, SCM s_callback)
{
  TessWrap	*wptr;
  int		 which, real_which;

  int		 index;
  SCM		 callbacks;
  CBFunc	 func;

  SCM_ASSERT (SCM_TESSP (s_tess), s_tess, SCM_ARG1, s_glu_tess_callback);
  SCM_ASSERT ((s_callback == SCM_BOOL_F) || (scm_procedure_p (s_callback) == SCM_BOOL_T),
	      s_callback, SCM_ARG3, s_glu_tess_callback);
  which = sgtk_scm2enum (s_which, &sgtk_glu_tess_callback_type_info, 
			 SCM_ARG2, (void *) s_glu_tess_callback);

  wptr = SCM_TESS (s_tess);
  callbacks = wptr -> callbacks;

  switch (which)
    {
    case GLU_TESS_BEGIN_DATA:
      callbacks = wptr -> callbacks_with_data;
      /* no break */
    case GLU_TESS_BEGIN:
      index = CB_BEGIN;
      real_which = GLU_TESS_BEGIN_DATA;
      break;
    case GLU_TESS_EDGE_FLAG_DATA:
      callbacks = wptr -> callbacks_with_data;
      /* no break */
    case GLU_TESS_EDGE_FLAG:
      index = CB_EDGE_FLAG;
      real_which = GLU_TESS_EDGE_FLAG_DATA;
      break;
    case GLU_TESS_VERTEX_DATA:
      callbacks = wptr -> callbacks_with_data;
      /* no break */
    case GLU_TESS_VERTEX:
      index = CB_VERTEX;
      real_which = GLU_TESS_VERTEX_DATA;
      break;
    case GLU_TESS_END_DATA:
      callbacks = wptr -> callbacks_with_data;
      /* no break */
    case GLU_TESS_END:
      index = CB_END;
      real_which = GLU_TESS_END_DATA;
      break;
    case GLU_TESS_COMBINE_DATA:
      callbacks = wptr -> callbacks_with_data;
      /* no break */
    case GLU_TESS_COMBINE:
      index = CB_COMBINE;
      real_which = GLU_TESS_COMBINE_DATA;
      break;
    case GLU_TESS_ERROR_DATA:
      callbacks = wptr -> callbacks_with_data;
      /* no break */
    case GLU_TESS_ERROR:
      index = CB_ERROR;
      real_which = GLU_TESS_ERROR_DATA;
      break;
    default:
      SCM_ASSERT (0, s_which, SCM_ARG2, s_glu_tess_callback);
    }
  
  SCM_VELTS (callbacks) [index] = s_callback;
  func = cb_func_array [index];

  if ((SCM_VELTS (wptr -> callbacks) [index] == SCM_BOOL_F) &&
      (SCM_VELTS (wptr -> callbacks_with_data) [index] == SCM_BOOL_F))
    func = NULL;

  gluTessCallback (wptr->tess, real_which, func);

  return SCM_UNSPECIFIED;
}

void sglu_init_GLU_tess_glue (void)
{
  tc_tess = scm_make_smob_type ("gluTesselator", sizeof (TessWrap));
  scm_set_smob_mark (tc_tess, smob_tess_mark);

#if ! (defined (SCM_MAGIC_SNARFER) || defined (SCM_MAGIC_SNARF_INITS))
#include "GLU-tess.x"
#endif
}

#endif
