; -*- Scheme -*-

; the main purpose of this file is debugging (rather than plain example)
; so excuse me if sometimes it is not straight and explanatory as it should be

(use-modules
 (gtk gdk)
 (gtk gtk)
 (open-gl gl)
 (open-gl glu)
 (open-gl gtkgl))

(define (Display . args)
  (for-each display args)
  (newline))

; for debugging - we do not want to lose list of views when reloading balls.scm

(if (not (defined? 'view-list))
    (define view-list '(#f)))

(define (feed-sphere-create wires)
  (let ((list-id		(gl-gen-lists 2)))

    (set! feed-sphere
	  (lambda (w) (gl-call-list (+ list-id (if w 1 0)))))

    (let
	((quadric		(glu-new-quadric))
	 (execute-wires		(if wires 'compile-and-execute 'compile))
	 (execute-fill		(if wires  'compile 'compile-and-execute)))
    
      (glu-quadric-draw-style quadric 'fill)
      (glu-quadric-normals quadric 'smooth)

      (gl-new-list list-id execute-fill)
        (glu-sphere quadric 1 16 8)
      (gl-end-list)

      (glu-quadric-draw-style quadric 'line)
      ; (glu-quadric-normals quadric 'none)

      (gl-new-list (+ list-id 1)  execute-wires)
        (glu-sphere quadric 1 16 8)
      (gl-end-list))))
    
(define feed-sphere feed-sphere-create)

(define (remove-glarea area)
  (set! view-list (delete area view-list))
  (cond ((equal? view-list '(#f))
	 (set! feed-sphere feed-sphere-create)
	 (gtk-main-quit))))

(define (setup-GL)
  (map gl-enable '(dither cull-face depth-test color-material))

  (gl-cull-face 'back)
  (gl-shade-model 'smooth)
  (gl-hint 'perspective-correction 'fastest)
  (gl-hint 'polygon-smooth 'fastest)
  (gl-color-material 'front 'ambient-and-diffuse)
  (gl-clear-color 0 0 0 1))

(define (create-new-view)
  (define timer		 (g-timer-new))
  (define time		 0)
  (define base-time	 0)

  (define fps		 20)

  (define wires		 #f)
  (define lights	 #t)

  (define fovy		 30)
  (define win-width	 #f)
  (define win-height	 #f)
  (define the-mover-id   #f)

  (define (update-base-time)
    (set! base-time (+ base-time (car (g-timer-elapsed timer)))))

  (define (setup-camera)
    (gl-matrix-mode 'projection)
    (gl-load-identity)
    (glu-perspective fovy (/ win-width win-height) 16 512)
    (gl-matrix-mode 'modelview))

  (define (setup-lights)
    (cond
     (lights
      (gl-light-modeli 'light-model-local-viewer 0)
      (gl-light-modeli 'light-model-two-side 0)

      (gl-enable 'light0)
      (gl-enable 'lighting)

      (gl-lightfv 'light0 'ambient #s(0.2 0.2 0.2 1))
      (gl-lightiv 'light0 'position #e(100 100 0 1)))
     (else
      (gl-disable 'light0)
      (gl-disable 'lighting))))

  (define (feed-scene)
    (gl-load-identity)
    (glu-look-at 100 100 100 
		 0 0 0 
		 0 1 0)

    (setup-lights)

    (gl-clear '(color-buffer depth-buffer))
  
    (gl-materialfv 'front 'diffuse #s( 0.6 0.6 0.6 1.0))
    (gl-materialfv 'front 'specular #s( 0.75 0.75 0.75 1.0))
    (gl-materialf 'front  'shininess 100)

    (gl-push-matrix)

    (gl-color3fv #s(1 1 0))
    (gl-scaled 10 10 10)
    (gl-enable 'normalize)
    (feed-sphere wires)

    (gl-pop-matrix)
    (gl-push-matrix)
    
    (gl-color3d 0 1 1)
    (gl-rotated (* 360 (/ time 10)) 2 -1 1)
    (gl-translated 0 45 0)
    (gl-scaled 3 3 3)
    (feed-sphere wires)

    (gl-pop-matrix)
    (gl-push-matrix)

    (gl-color3d 0 1 0)
    (gl-rotated (* 360 (/ time 7)) 2 -1 1)
    (gl-translated 0 0 33)
    (gl-scaled 2 2 2)
    (feed-sphere wires)

    (gl-pop-matrix)
    (gl-disable 'normalize)

    (gl-rotated (* 360 (/ time -3)) 0 1 5)
    (gl-translated 25 0 0)
    (gl-color3d 1 0 0)
    (feed-sphere wires))

  (let*
      ((window	(gtk-window-new 'toplevel))
       (vbox	(gtk-vbox-new #f 5))

       (hbox	(gtk-hbox-new #f 5))
       (label	(gtk-label-new "Required fps: "))
       (adjust  (gtk-adjustment-new 20 1 100 1 1 0))
       (spinner	(gtk-spin-button-new adjust 1 0))
       
       (frame	(gtk-frame-new #f))
       (glarea  (gtk-gl-area-new '(use-gl rgba doublebuffer depth-size 2) (car view-list)))
       (new-win	(gtk-button-new-with-label "New window"))

       (hbox2	(gtk-hbox-new #f 2))
       (t-spin	(gtk-toggle-button-new-with-label "Spin"))
       (t-light	(gtk-toggle-button-new-with-label "Light"))
       (t-wires (gtk-toggle-button-new-with-label "Wires"))

       (reload	(gtk-button-new-with-label "Reload balls.scm"))
       (run-gc	(gtk-button-new-with-label "Run (gc)"))
       (close	(gtk-button-new-with-label "Close"))
       (quit	(gtk-button-new-with-label "Quit"))

       (render-scene (lambda () (with-gtk-gl-area:render! glarea (feed-scene))))
       (the-mover (lambda () 
		    (set! time (+ base-time  (car (g-timer-elapsed timer))))
		    (render-scene)
		    (do () ((eq? 0 (gtk-events-pending))) (gtk-main-iteration))
		    #t))
       (timeout  #f)
       (new-timeout (lambda () (gtk-timeout-add (inexact->exact (floor (/ 1000 fps))) the-mover))))
      
    (gtk-window-set-title window "GuileGL: balls")
    (gtk-container-border-width window 5)
    (gtk-container-add window vbox)

    (gtk-widget-add-events window '(key-press-mask key-release-mask))

    ;;; Layout of window

    (gtk-box-pack-start vbox hbox #f #f 0)
      (gtk-box-pack-start hbox label #f #f 0)
      (gtk-box-pack-end hbox spinner #f #f 0)
      (gtk-widget-set-usize spinner 50 0)

    (gtk-box-pack-start vbox frame #t #t 10)
      (gtk-container-add frame glarea)
      (gtk-widget-set-usize glarea 150 150)

    (for-each (lambda (x) (gtk-box-pack-start hbox2 x #t #t 0))
	      (list t-spin t-light t-wires))

    (for-each (lambda (x) (gtk-box-pack-start vbox x #f #f 0))
	      (list hbox2 new-win reload run-gc close quit))

    ;;; setup various things...

    (set! view-list (cons glarea view-list))

;   (gtk-toggle-button-set-state t-spin #f)
    (gtk-toggle-button-set-state t-light #t)
;   (gtk-toggle-button-set-state t-wires #f)

    ;;; callbacks

    (gtk-signal-connect new-win "clicked" create-new-view)
    (gtk-signal-connect reload "clicked" (lambda () (load "balls.scm")))
    (gtk-signal-connect run-gc "clicked" gc)
    (gtk-signal-connect close "clicked" (lambda () (gtk-widget-destroy window)))
    (gtk-signal-connect quit "clicked" gtk-main-quit)

    (gtk-signal-connect window "destroy" (lambda () (remove-glarea glarea)))
    (gtk-signal-connect glarea "destroy" (lambda () (if timeout (gtk-timeout-remove timeout))))

    (gtk-signal-connect glarea "expose_event"
			(lambda data (render-scene)))

    (gtk-signal-connect glarea "configure_event"
			(lambda (e)
			  (set! win-width (gdk-event-configure-width e))
			  (set! win-height (gdk-event-configure-height e))

			  (with-gtk-gl-area glarea
			    (gl-viewport 0 0 win-width win-height)
			    (setup-camera))))

    (gtk-signal-connect glarea "realize"
			(lambda data
			  (with-gtk-gl-area glarea
			    (setup-GL))))

    (gtk-signal-connect window "key_press_event"
			(lambda (data)
			  (let* ((keyval	(gdk-event-keyval data))
			  	 (key (if (< keyval 255) (integer->char keyval) #f)))
			    (case key
			      ((#\+ #\- #\=)
			       (gtk-signal-emit-stop-by-name window "key_press_event")
			       (set! fovy ((if (eqv? key #\-) + -) fovy 5))
			       (set! fovy
				     (cond
				      ((< fovy 10) 10)
				      ((> fovy 120) 120)
				      (else fovy)))
		
			       (with-gtk-gl-area:render! glarea
				  (setup-camera)
				  (feed-scene)))))))

    (gtk-signal-connect t-light "toggled"
			(lambda data
			  (set! lights (gtk-toggle-button-active t-light))
			  (render-scene)))

    (gtk-signal-connect t-wires "toggled"
			(lambda data
			  (set! wires (gtk-toggle-button-active t-wires))
			  (render-scene)))

    (gtk-signal-connect t-spin "toggled"
			(lambda data
			  (cond
			   ((gtk-toggle-button-active t-spin)
			    (g-timer-reset timer)
			    (g-timer-start timer)
			    (set! timeout (new-timeout)))
			   (else
			    (g-timer-stop timer)
			    (update-base-time)
			    (gtk-timeout-remove timeout)
			    (set! timeout #f)))))

    (gtk-signal-connect adjust "value-changed"
			(lambda data
			  (let ((newfps	(gtk-spin-button-get-value-as-int spinner)))
			    (cond 
			     ((not (eqv? newfps fps))
			      (set! fps newfps)

			      (g-timer-stop timer)
			      (update-base-time)

			      (gtk-timeout-remove timeout)
			      (set! timeout (new-timeout))

			      (g-timer-reset timer)
			      (g-timer-start timer))))))
    (gtk-widget-show-all window)))


(create-new-view)
(and 
 (gtk-standalone?)
 (eq? (gtk-main-level) 0)
 (gtk-main))
