(use-modules (open-gl	glu)
	     (open-gl	gl)
	     (open-gl	simple-gl-window))


(define nof-spheres 4)
(define spheres (make-vector 4 #f))

(define quadric (glu-new-quadric))

(glu-quadric-draw-style quadric 'fill)
(glu-quadric-normals quadric 'smooth)

(define sphere-params
;	 r  p1 p2  r-coeff    r-axe    i-translate color
  '((0	(10 16 8)  (  0	  1) (0  1  0) ( 0  0  0) (1 1 0))
    (1  ( 3  8 4)  (180	  5) (2 -1  1) ( 0 45  0) (0 1 1))
    (2  ( 2  8 4)  (360	  7) (2 -1 -1) ( 0  0 33) (0 1 0))
    (3  ( 1  4 2) (-360	  3) (0  1  5) (25  0  0) (1 0 0))))

(define (render-sphere id)
  (let ((list-id	(vector-ref spheres id)))
    (if list-id 
	(gl-call-list list-id)
	(let ((new-list-id	(gl-gen-lists 1))
	      (params		(assoc-ref sphere-params id)))
	  (vector-set! spheres id new-list-id)
	  (gl-new-list new-list-id 'compile-and-execute)
	    (apply glu-sphere (cons quadric (car params)))
	  (gl-end-list)))))


(define (render-scene time)
  (gl-clear '(color-buffer depth-buffer))

  (gl-load-identity)
  (glu-look-at 100 100 100
	         0   0   0
		 0   1   0)

  (for-each 
   (lambda (x)
     (apply (lambda (id _ign r-coeff r-axe i-translate color)
	      (gl-color3dv color)
	      (gl-push-matrix)
	      (apply gl-rotated 
		     (cons (* (car r-coeff) 
			      (/ time (cadr r-coeff)))
			   r-axe))
	      (apply gl-translated i-translate)
	      (render-sphere id)
	      (gl-pop-matrix))
	    x))
   sphere-params))
	      

(define (setup-GL)
  (for-each gl-enable '(cull-face depth-test color-material))
  (gl-cull-face 'back)
  (gl-shade-model 'smooth)

	; hints
  (gl-hint 'perspective-correction 'fastest)
  (gl-hint 'polygon-smooth 'fastest)

	; lights
  (gl-enable 'light0)
  (gl-enable 'lighting)

  (gl-lightfv 'light0 'ambient #(0.2 0.2 0.2 1))
  (gl-lightfv 'light0 'position  #(100 100 0 1))

	; colors and materials
  (gl-clear-color 0 0 0 1)
  (gl-color-material 'front 'ambient-and-diffuse))

(define (my-enable what enable)
  ((if enable gl-enable gl-disable) what))

(define (configure-callback context fovy width height)
  (gl-viewport 0 0 width height)
  (gl-matrix-mode 'projection)
  (gl-load-identity)
  (glu-perspective fovy (/ width height) 16 512)
  (gl-matrix-mode 'modelview))

(define (update-fovy context fovy step)
  (let* ((temp		(+ fovy step))
	 (newfovy	(max 5 (min temp 120))))
    (configure-callback context newfovy
			(simple-gl-width context)
			(simple-gl-height context))
    newfovy))

(define (new-window)
  (let* ((fovy		45))
    (simple-gl-context 

	; Name of the game

		"Balls animation"

	; expose event

		(lambda (context)
		  (render-scene (simple-gl-time-elapsed context)))

	; realize event

		(lambda (context)
		  (setup-GL))

	; configure-event

		'configure 

		(lambda (context width height)
		  (configure-callback context fovy width height))

	; clone 

		'clone

		(lambda (context)
		  (new-window))

	; initial size

		'size	'(200 200)

	; GL attribute list

		'attribute-list '(rgba doublebuffer depth-size 2)

	; menus

		'menu

		`((default)
		  (---> "_Render"
			((*---)
			 (X... "_Lighting" 
			       key "<Ctrl>L" 
			       init-state #t
			      ,(lambda (context on)
				 (with-simple-gl-context:render! context
				   (my-enable 'lighting on))))
			 (X... "_Depth test"
			       key "<Ctrl>D"
			       init-state #t
			      ,(lambda (context on)
				 (with-simple-gl-context:render! context
			           (my-enable 'depth-test on))))
			 (X... "_Cull face"
			       key "<Ctrl>C"
			      ,(lambda (context on)
				 (with-simple-gl-context:render! context
			           (my-enable 'cull-face on))))
			 (X... "_Wires"
			       key "<Ctrl>i"
			       init-state #f
			      ,(lambda (context on)
				 (with-simple-gl-context:render! context
				   (if on
				     (gl-polygon-mode 'front-and-back 'line)
				     (gl-polygon-mode 'front 'fill)))))
			 (----)
			 (X... "_Animation"
			       key "<Ctrl>a"
			       init-state #f
			      ,(lambda (context on)
				 (if on
				     (simple-gl-timer-start 
				        context #t
					(lambda (context)
					  (simple-gl-redraw context)))
				     (simple-gl-timer-stop context))))))

		  (---> "_Zoom"
			((*---)
			 ("Zoom _in" key "="
				    ,(lambda (context) 
				       (with-simple-gl-context:render! context
				          (set! fovy (update-fovy context fovy  -5)))))

			 ("Zoom _out" key "-"
				    ,(lambda (context) 
				       (with-simple-gl-context:render! context
				          (set! fovy (update-fovy context fovy 5)))))))
		  ))))

(new-window)
(simple-gl-run)