(use-modules
 (open-gl glu)
 (open-gl gl)
 (open-gl simple-gl-window))

(define quadric (glu-new-quadric))
(glu-quadric-draw-style quadric 'fill)

	; select buffer have to be an uniform array of unsigned integers

(define num-of-objects 3)
(define select-buffer (make-uniform-array 1 (* 4 num-of-objects)))

(define object-names
  '((1	"triangle")
    (2	"ring")
    (3  "rectangle")))

(define (display-selected-iter n)
  (let* ((index	 (+ 3 (* 4 (- n 1))))
	 (name   (assq-ref object-names (array-ref select-buffer index))))

    (if (eq? n 1)
	name
	(cons (car name) (cons ", " (display-selected-iter (- n 1)))))))

(define (display-selected context n)
  (simple-gl-Display context "Selected objects: "
		     (if  (eq? n 0) 
			  "none"
			  (list->string (display-selected-iter n)))))

(define (Welcome context)
  (simple-gl-Display context "Welcome to simple-select demo...\n")
  (simple-gl-Display context "Please select object with mouse."))

(define (draw-scene)

  (gl-push-name 3)
  (gl-color3d .5 .8 1)
  
  (with-gl-begin 'quads
    (gl-vertex2d 55 50)
    (gl-vertex2d 95 50)
    (gl-vertex2d 95 90)
    (gl-vertex2d 55 90))

  (gl-load-name 2)
  (gl-color3d .9 .9 0.5)

  (gl-push-matrix)
  (gl-translated 50 45 0)

  (glu-disk quadric 13 25 24 1)

  (gl-pop-matrix)

  (gl-load-name 1)
  (gl-color3d 1 .4 .4)

  (with-gl-begin 'triangles
    (gl-vertex2dv #(05 50))
    (gl-vertex2dv #(45 50))
    (gl-vertex2dv #(25 90)))

  (gl-pop-name))

(define (setup-projection-matrix context . pick)
  (gl-matrix-mode 'projection)
  (gl-load-identity)

  (if (null? pick)
      (glu-ortho2d 0 100 0 100)
      (let* ((width	(simple-gl-width context))
	     (height	(simple-gl-height context))
	     (x		(* 100 (/ (car pick) width)))
	     (y		(- 100 (* 100 (/ (cadr pick) height))))
	     (delta	0.1))

	(glu-ortho2d (- x delta) (+ x delta) (- y delta) (+ y delta))))

  (gl-matrix-mode 'modelview))

(simple-gl-context "Select"

	; expose (= redraw) event

		    (lambda (context)
		      (gl-clear '(color-buffer))
		      (draw-scene))

	; realize (= setup) event

		    (lambda (context)
		      (gl-select-buffer select-buffer)
		      (gl-clear-color 0 0 0 1)

		      (setup-projection-matrix context) 
		      (gl-load-identity)
		      
		      (Welcome context))

	; mouse click event

		  'mouse-click (lambda (context x y)

			; setup pick matrix				 
				 (setup-projection-matrix context x y)

			; setup select mode
				 (gl-render-mode 'select)
				 (gl-init-names)

			; redraw scene in select mode				  
				   (draw-scene)

			; shut down select mode and display selected lists				   
				  (display-selected context (gl-render-mode 'render))

			; restore projection matrix				
				  (setup-projection-matrix context))


	; turn on report window and set its height to 80 pixels
		  'report 80

	; initial size
		  'size '(300 300)

	; enable menu (at least - define just simple menu)
		  'menu		'((default-menu)))

(simple-gl-run)