#include <gtk/gtk.h>
#include <gtkgl/gdkgl.h>
#include <gtkgl/gtkglarea.h>

#include <libguile.h>
#include <guile/gh.h>
#include <guile-gtk.h>
#include <string.h>

typedef struct {
  char*		name;
  int		glx_value;
  int		intparam;
  int		name_len;
} sgtk_glXCC_type;

extern void sgtk_init_gtkgl_glue_from_defs (void);

static sgtk_glXCC_type sgtk_glXCC_params [] = 
{
  {"use-gl",		GDK_GL_USE_GL,			0, -1},
  {"buffer-size",	GDK_GL_BUFFER_SIZE,		1, -1},
  {"level",		GDK_GL_LEVEL,			1, -1},
  {"rgba",		GDK_GL_RGBA,			0, -1},
  {"doublebuffer",	GDK_GL_DOUBLEBUFFER,		0, -1},
  {"stereo",		GDK_GL_STEREO,			0, -1},
  {"aux-buffers",	GDK_GL_AUX_BUFFERS,		1, -1},
  {"red-size",		GDK_GL_RED_SIZE,		1, -1},
  {"green-size",	GDK_GL_GREEN_SIZE,		1, -1},
  {"blue-size",		GDK_GL_BLUE_SIZE,		1, -1},
  {"alpha-size",	GDK_GL_ALPHA_SIZE,		1, -1},
  {"depth-size",	GDK_GL_DEPTH_SIZE,		1, -1},
  {"stencil-size",	GDK_GL_STENCIL_SIZE,		1, -1},
  {"accum-red-size",	GDK_GL_ACCUM_RED_SIZE,		1, -1},
  {"accum-green-size",	GDK_GL_ACCUM_GREEN_SIZE,	1, -1},
  {"accum-blue-size",	GDK_GL_ACCUM_BLUE_SIZE,		1, -1},
  {"accum-alpha-size",	GDK_GL_ACCUM_ALPHA_SIZE,	1, -1},
  {NULL,-1,-1,-1}
};

static int*
sgtk_parse_glXChooseVisual_params (SCM s_attrlist, const char *proc)
{
  int		*attrlist, *aptr;
  int		size;

  int		symlen;
  char		*symbol;

  SCM		s_head, s_value, s_err;
  sgtk_glXCC_type *ptr;

  size = scm_ilength (s_attrlist);
  aptr = attrlist = g_new (int, size + 1);

  s_head = s_attrlist;
  while (s_head != SCM_EOL)
    {
      s_err = s_head;
      s_value = SCM_CAR (s_head);
      s_head = SCM_CDR (s_head);

      if (! SCM_SYMBOLP (s_value))
	goto error;

      symbol = SCM_CHARS (s_value);
      symlen = SCM_LENGTH (s_value);

      for (ptr = sgtk_glXCC_params; 
	   !((ptr->name_len == symlen) && (strncmp (ptr->name, symbol, symlen) == 0));
	   ptr++)
	{
	  if (ptr->name == NULL)
	    goto error;
	}
      
      *aptr++ = ptr -> glx_value;

      if (ptr->intparam)
	if (s_head == SCM_EOL)
	  goto error;	  
	else	
	  {
	    s_value = SCM_CAR (s_head);
	    s_head = SCM_CDR (s_head);

	    if (! SCM_INUMP (s_value))
	      goto error;
	    
	    *aptr++ = gh_scm2int (s_value);
	  }
    }

  *aptr = GDK_GL_NONE;
  return attrlist;

 error:
  g_free (attrlist);
  SCM_ASSERT (0, s_err, SCM_ARGn, proc);

  return NULL;
}


SCM_PROC (s_gtk_gl_area_new, "gtk-gl-area-new", 1, 1, 0, sgtk_gl_area_new);

static SCM
sgtk_gl_area_new (SCM s_attrlist, SCM s_share)
{
  GtkWidget	*share;
  GtkWidget	*result;
  int		*attrlist;

  SCM_ASSERT ((s_share == SCM_UNDEFINED) || (s_share == SCM_BOOL_F) 
	      || sgtk_is_a_gtkobj (gtk_gl_area_get_type (), s_share),
	      s_share, SCM_ARG1, s_gtk_gl_area_new);

  share = ((s_share == SCM_UNDEFINED) || (s_share == SCM_BOOL_F)) ? NULL 
    : (GtkWidget *) sgtk_get_gtkobj (s_share);
  attrlist = sgtk_parse_glXChooseVisual_params (s_attrlist, s_gtk_gl_area_new);
  
  SCM_DEFER_INTS;
  result = gtk_gl_area_share_new (attrlist, (GtkGLArea *) share);
  g_free (attrlist);
  SCM_ALLOW_INTS;

  return sgtk_wrap_gtkobj ((GtkObject *) result);
}

void sgtk_init_gtkgl_support_glue (void)
{
  sgtk_glXCC_type *ptr;

  for (ptr = sgtk_glXCC_params; ptr->name != NULL; ptr++)
    ptr->name_len = strlen (ptr->name);

#if ! (defined (SCM_MAGIC_SNARFER) || defined (SCM_MAGIC_SNARF_INITS))
#include "gtkgl-support.x"
#endif
}
