; -*- Scheme -*-

(define-module (open-gl simple-gl-window)
 :use-module (gtk gtk)
 :use-module (gtk gdk)
 :use-module (open-gl gl)
 :use-module (open-gl glu)
 :use-module (open-gl gtkgl))

(define-public context-list '())
(define-public GL-shared-list '())

(define (get-param param-list name default)
  (cond
   ((null? param-list)
    default)
   ((eq? name (car param-list))
    (cadr param-list))
   (else
    (get-param (cddr param-list) name default))))


(define (extract-opt-parameters parameters)
  (cdddr parameters))

(define (get-menu-param param-list name is-a? default)
  (cond
   ((null? param-list)
    default)
   ((eq? name (car param-list))
    (cadr param-list))
   ((and is-a? (is-a? (car param-list)))
    (car param-list))
   (else
    (get-menu-param (cdr param-list) name is-a? default))))

(define (create-item create-func name parent accel-group hotkey)
  (let* ((item	(create-func))
	 (label	(gtk-accel-label-new ""))
	 (u-key	(gtk-label-parse-uline label (stringify name)))
	 (a-key	(and (string? hotkey)
		     (or
		      (and (eq? 1 (string-length hotkey))
			   (cons (char->integer (string-ref hotkey 0)) '(())))
		      (gtk-accelerator-parse hotkey)))))
    
    (gtk-misc-set-alignment label 0 0.5)
    (gtk-container-add item label)
    (gtk-accel-label-set-accel-widget label item)

    (if a-key 
	(gtk-accel-group-add accel-group (car a-key) (cadr a-key) '(visible) item "activate"))
    (or (eq? u-key 16777215)   ; GDK_VoidSymbol
	(if (gtk-menu-bar? parent)
	    (gtk-widget-add-accelerator item "activate_item" accel-group
					u-key '(mod1-mask) '(locked))
	    (gtk-widget-add-accelerator item "activate_item"
					(gtk-menu-ensure-uline-accel-group parent)
					u-key '() '(locked))))
    item))

(define (process-menu context menu-list container append-func accel-group)
  (if (pair? menu-list)
      (let* ((desc	(car menu-list))
	     (id	(car desc))
	     (param	(cdr desc))
	     (rest	(cdr menu-list)))
	(case id
	  ((--- ---- ----- separator)
	   (let ((w (gtk-menu-item-new)))
	     (gtk-widget-set-sensitive w #f)
	     (append-func container w)))
	  ((*-- *--- *---- tear-off)
	   (let ((w (gtk-tearoff-menu-item-new)))
	     (append-func container w)))
	  ((default default-menu)
	   (process-menu context
			 `((--> "_GuileGL"
				((*---)
				 ("_New window" key "<Ctrl>n" ,(lambda (w) (clone-context w)))
				 ("_Close window" key "<Ctrl>w" ,(lambda (w) (simple-gl-close-window w)))
				 (----)
				 ("E_xit" key "<Ctrl>q" ,(lambda (w) (gtk-main-quit))))))
			 container append-func accel-group))
	  ((--> ---> ----> submenu)
	   (let* ((name  (cadr desc))
		  (submenu (get-menu-param param #f list? '()))

		  (item   (create-item gtk-menu-item-new name container accel-group #f))
		  (menu   (gtk-menu-new)))

	     (append-func container item)
	     (gtk-menu-item-set-submenu item menu)

	     (process-menu context submenu menu gtk-menu-append accel-group)))
	  ((X.. X... X.... radio-item)
	   (let* ((name		(cadr desc))
		  (hotkey	(get-menu-param param 'key #f #f))
		  (widget	(create-item gtk-check-menu-item-new name container accel-group hotkey))
		  (func		(get-menu-param param #f procedure? #f))
		  (init		(get-menu-param param 'init-state #f #f)))

	     (append-func container widget)
	     (gtk-check-menu-item-set-state widget init)
	     (if func 
		 (gtk-signal-connect widget "toggled" 
				     (lambda ()
				       (func context (gtk-check-menu-item-active widget)))))))

	  (else  ; boring menu item
	   (let* ((func   (get-menu-param param #f procedure? #f))
		  (hotkey (get-menu-param param 'key #f #f))
		  (widget (create-item gtk-menu-item-new id container accel-group hotkey)))

	     (if func (gtk-signal-connect widget "activate" (lambda () (func context))))
	     (append-func container widget))))

	(process-menu context rest container append-func accel-group))))

(define-public (simple-gl-context name expose realize . other)
  (let* ((shared	(get-param other 'shared #t))
	 (shared-widget (and shared (pair? GL-shared-list) (car GL-shared-list)))

	 (size		(get-param other 'size  '(100 100)))
	 (click-event	(get-param other 'mouse-click #f))
	 (report	(get-param other 'report #f))
	 (attrib-list	(get-param other 'attribute-list '(rgba doublebuffer)))


	 (window	(gtk-window-new 'toplevel))
	 (vbox		(gtk-vbox-new #f 5))
	 (gla		(gtk-gl-area-new attrib-list shared-widget))

	 (conf-event	(get-param other 
				   'configure  
				   (lambda (context width height)
				     (gl-viewport 0 0 width height))))

	 (menu-list	(get-param other 'menu '()))
	 (context	`((GL-area ,gla) (time-info 0 #f) (window ,window) 
			  (parameters ,@(append (list name expose realize) other))))
	 (expose-cb	(lambda data (with-gtk-gl-area:render! gla (expose context)))))

    (or gla (error "Can not create GL context"))

    (set! context (cons (list 'expose expose-cb) context))
    
    (gtk-window-set-title window (string-append name ": simple guileGL window"))
    (gtk-signal-connect window "destroy" (lambda () (destroy-callback context)))
 
    (gtk-container-border-width vbox 10)
    (gtk-box-pack-start vbox gla #t #t 0)
    (gtk-widget-add-events gla '(exposure-mask))

    (gtk-signal-connect gla "expose_event" expose-cb)
    (gtk-signal-connect gla "configure_event"
			(lambda (event)
			  (with-gtk-gl-area gla
			     (conf-event context
					 (gdk-event-configure-width event)
					 (gdk-event-configure-height event)))))

    (gtk-signal-connect gla "realize"
			(lambda data
			  (with-gtk-gl-area gla (realize context))))

    (if click-event
	(begin
	  (gtk-widget-add-events gla '(button-press-mask))
	  (gtk-signal-connect gla "button_press_event"
			      (lambda (event)
				(if (eq? (gdk-event-type event) 'button-press)
				    (with-gtk-gl-area gla
				       (click-event context
						    (gdk-event-x event)
						    (gdk-event-y event))))))))

    (if report
	(let* ((adjust	(gtk-adjustment-new 10 0 10 0.1 1.0 1))
	       (hbox	(gtk-hbox-new #f 0))
	       (txtout	(gtk-text-new #f adjust))
	       (scroll	(gtk-vscrollbar-new adjust)))

	  (gtk-box-pack-start hbox txtout #t #t 0)
	  (gtk-box-pack-end hbox scroll #f #t 0)

	  (gtk-widget-set-usize txtout 0 (if (number? report) report 50))
	  (gtk-box-pack-start vbox hbox #f #f 0)

	  (append! context `((report ,txtout ,adjust)))))

    (if (null? menu-list)
	(gtk-container-add window vbox)
	(let* ((outer-vbox	(gtk-vbox-new #f 0))
	       (menu-bar	(gtk-menu-bar-new))
	       (accel-group	(gtk-accel-group-new)))

	  (gtk-accel-group-attach accel-group window)
	  (gtk-container-add window outer-vbox)

	  (gtk-box-pack-start outer-vbox menu-bar #f #f 0)
	  (gtk-box-pack-start outer-vbox vbox #t #t 0)

	  (process-menu context menu-list menu-bar gtk-menu-bar-append accel-group)))

    (gtk-widget-set-usize gla (car size) (cadr size))
    (gtk-widget-show-all window)

    (set! context-list (cons context context-list))
    (if shared 
	(set! GL-shared-list (cons gla GL-shared-list)))

    context))

(define-public (simple-gl-width context)
  (gtk-widget-allocation-width (car (assoc-ref context 'GL-area))))

(define-public (simple-gl-height context)
  (gtk-widget-allocation-height (car (assoc-ref context 'GL-area))))

(define (clone-context context)
  (let* ((parameters (assoc-ref context 'parameters))
	 (clone	     (get-param (extract-opt-parameters parameters) 'clone #f)))

    (if clone 
	(clone context)
	(apply simple-gl-context parameters))))

(define-public (simple-gl-close-window context)
  (gtk-widget-destroy (car (assoc-ref context 'window))))

(define (destroy-callback context)
  (let	((window	(car (assq-ref context 'window)))
	 (gla		(car (assq-ref context 'GL-area))))

    (main-timeout-remove-context context)

    (set! context-list (delete context context-list))
    (set! GL-shared-list (delete gla GL-shared-list))

    (gtk-widget-destroy window)
    (if (null? context-list) (gtk-main-quit))))

;;; various usefull macros and functions

(define-public (simple-gl-make-current context)
  (let ((gla	(car (assoc-ref context 'GL-area))))
    (gtk-gl-area-make-current gla)))

(defmacro-public with-simple-gl-context (context . body)
  `(begin
     (simple-gl-make-current ,context)
     ,@body))

(defmacro-public with-simple-gl-context:render! (context . body)
  `(begin
     (simple-gl-make-current ,context)
     ,@body
     (simple-gl-redraw context)))
  
(define-public (simple-gl-redraw context)
  (let ((expose (car (assoc-ref context 'expose))))
    (expose context)))


;;; timers

(define (clear-gtk-queue)
  (cond 
   ((< 0 (gtk-events-pending))
    (gtk-main-iteration)
    (clear-gtk-queue))))

(define main-timeout-cps	30)
(define main-timeout-call-list	'())
(define main-timeout-id		#f)
(define main-timeout-dirty-flag #f)

(define (main-timeout-iter list)
  (and (pair? list)
       (begin
	 (and (car list) ((cadar list) (caar list)))
	 (clear-gtk-queue)
	 (main-timeout-iter (cdr list)))))
      
(define (main-timeout-callback)
  (cond
   (main-timeout-dirty-flag
    (set! main-timeout-dirty-flag #f)
    (set! main-timeout-call-list
	  (main-timeout-clear-unused main-timeout-call-list))))
  (cond
   ((null? main-timeout-call-list)
    (set! main-timeout-id #f)
    #f)		; #f = remove callback
   (else
    (main-timeout-iter main-timeout-call-list)
    #t)))

(define (main-timeout-remove-context-from-list list context)
  (cond 
   ((pair? list)
    (and (equal? (caar list) context)
	 (set-car! list #f))
    (main-timeout-remove-context-from-list (cdr list) context))))

(define (main-timeout-remove-context context)
  (set! main-timeout-dirty-flag #t)
  (main-timeout-remove-context-from-list main-timeout-call-list context))

(define (main-timeout-add-context context callback)
  (set! main-timeout-call-list
	(cons (list context callback)
	      main-timeout-call-list))
  (or main-timeout-id
      (set! main-timeout-id 
	    (gtk-timeout-add (inexact->exact (floor (/ 1000 main-timeout-cps)))
			     main-timeout-callback))))

(define (main-timeout-clear-unused list)
  (cond
   ((null? list) '())
   ((not (car list))
    (main-timeout-clear-unused (cdr list)))
   (else
    (cons (car list) (main-timeout-clear-unused (cdr list))))))
    
(define-public (simple-gl-fps fps)
  (cond
   ((not (eq? fps main-timeout-fps))
    (set! main-timeout-fps fps)
    (cond 
     (main-timeout-id
      (and main-timeout-id (gtk-timeout-remove main-timeout-id))
      (set! main-timeout-id 
	    (gtk-timeout-add (floor (/ 1000 fps)) 
			     main-timeout-callback)))))))


;;; simple-gl api    

(define (get-time-info context)
  (assq-ref context 'time-info))

(define-public (simple-gl-time-elapsed context)
  (let* ((time-info	(get-time-info context))
	 (timer		(cadr time-info)))
    (+ (car time-info) 
       (if timer
	   (car (g-timer-elapsed timer))
	   0))))

(define-public (simple-gl-timer-stop context)
  (let*  ((time-info	(get-time-info context))
	  (g-timer-pair		(cdr time-info)))

    (cond 
     ((car g-timer-pair)
      (main-timeout-remove-context context)
      (set-car! time-info (+ (car time-info)
			     (car (g-timer-elapsed (car g-timer-pair)))))))))


(define-public (simple-gl-timer-start context reset-time proc)
  (let* ((time-info     	(get-time-info context))
	 (g-timer-pair		(cdr time-info)))

    (and reset-time (set-car! time-info 0))

    (if (car g-timer-pair)
	(g-timer-reset (car g-timer-pair))
	(set-car! g-timer-pair (g-timer-new)))

    (main-timeout-add-context context proc)))

;;; display functions

(define-public (simple-gl-run) 
  (and (gtk-standalone?) (gtk-main)))


(define (Display . x)
  (for-each display x) (newline))

(define (stringify x)
  (cond
   ((string? x) x)
   ((number? x)
    (string->number x))
   ((symbol? x)
    (symbol->string x))
   (else
    (call-with-output-string (lambda (port) (display x port))))))

(define (text-widget-write tw x)
  (gtk-text-insert tw #f #f #f (stringify x) -1))

(define (adjust-set-end adj)
  (gtk-adjustment-set-value adj (- (gtk-adjustment-upper adj) (gtk-adjustment-page-size adj))))

(define-public (simple-gl-display context message)
  (let ((report		(assq-ref context 'report)))
    (if report
	(let ((txtout	(car report))
	      (adjust	(cadr report)))

	  (gtk-text-freeze txtout)
	  (text-widget-write txtout message)
	  (gtk-text-thaw txtout)
	  (adjust-set-end adjust))

	(display message))))

(define-public (simple-gl-newline x)
  (simple-gl-display x "\n"))

(define-public (simple-gl-Display context . messages)
  (let ((report		(assoc-ref context 'report)))
    (if report
	(let ((txtout	(car report))
	      (adjust	(cadr report)))

	  (gtk-text-freeze txtout)
	    (for-each (lambda (s) (text-widget-write txtout s)) messages)
	    (text-widget-write txtout "\n")
	  (gtk-text-thaw txtout)
	  (adjust-set-end adjust))

	(apply Display messages))))
